#!/usr/bin/perl
#
#  Copyright 2000, G. Wade Johnson
#  Released under the same license as Perl.
#  
#  Version: 0.7.1
#
#  This module provides support code for dealing with PLL structures in
#  memory.
#

package PLL;

use WSDL;
use XML::Data::DataDoc;

use strict;

use constant DEFAULT_GUTTER_WIDTH => 5;

my %LegalTags =
  (
    layout  => { attr => { class => 0 },
                 content => { prelayout => 0, head => 0, body => 1 },
                 exec => \&perform_layout
               },
    prelayout => { attr => { ref => 0 },
                 content => { },
                 exec => sub { } 
               },
    head    => { attr => { },
                 content => { text => 0, code => 0 },
                 exec => \&layout_head
               },
    body    => { attr => { cols => 0, rows => 0 },
                 content => { row => 0, content => 0, header => 0,
                              footer => 0, text => 0, code => 0 },
                 exec => \&layout_body
               },
    row     => { attr => { class => 0 },
                 content => { cell => 0, gutter => 0 },
                 exec => \&layout_row
               },
    cell    => { attr => { width => 0, colspan => 0, rowspan => 0, class => 0,
                           valign => 0, align => 0 },
                 content => { header => 0, footer => 0, code => 0,
                              text => 0, content => 0 },
                 exec => \&layout_cell
               },
    gutter  => { attr => { width => 0, rowspan => 0, class => 0 },
                 content => { }, 
                 exec => \&layout_gutter
               },
    header  => { attr => { 'ref' => 0 }, 
                 content => { }, 
                 exec => sub { layout_element( 'header', @_ ); }
               },
    footer  => { attr => { 'ref' => 0 },
                 content => { },
                 exec => sub { layout_element( 'footer', @_ ); }
               },
    code    => { attr => { 'ref' => 1 }, 
                 content => { }, 
                 exec => \&layout_code
               },
    text    => { attr => { 'ref' => 1 },
                 content => { },
                 exec => sub { layout_element( 'text', @_ ); }
               },
    content => { attr => { 'file' => 0 }, 
                 content => { },
                 exec => \&layout_content
               }
  );


#----------------------------------------------------------
#  Validate a pll structure 
sub   validate
 {
  my $pll  = shift;
  my $wsdl = shift;
  my $type = $pll->Type();

  die "Invalid root element \'$type\'.\n" unless 'layout' eq $type;
  _validate( $pll, $wsdl );

  # test for proper rows and columns if specified.
  _validate_body_table( $pll );
}



#----------------------------------------------------------
#  Validate a pll element
sub   _validate
 {
  my $elem = shift;
  my $wsdl = shift;
  my $type = $elem->Type();
  my $desc = $LegalTags{$type};

  die "Tag $type is not recognized.\n" unless defined $desc;

  # test attributes
  my $at;
  foreach $at ($elem->AttribNames())
   {
    die "Unrecognized attribute \'$at\' on element \'$type\'.\n"
                                         unless exists $desc->{attr}->{$at};
   }

  # test for required attributes
  foreach $at (grep { $desc->{attr}->{$_} } keys %{$desc->{attr}})
   {
    die "Missing attribute \'$at\' on element \'$type\'.\n"
                                         unless $elem->HasAttrib($at);
   }

  # test content
  my $el;
  foreach $el ($elem->ElementContent())
   {
    my $eltype = $el->Type();
    die "Unexpected element type \'$eltype\' in content for \'$type\'.\n"
                                        unless exists $desc->{content}->{$eltype};
    _validate( $el, $wsdl );
   }

  foreach $el (grep { $desc->{content}->{$_} } keys %{$desc->{content}})
   {
    die "Missing element type \'$el\' in content for \'$type\'.\n"
                                                unless $elem->Content( $el );
   }
 }


#----------------------------------------------------------
#  Validate the constraints on the body table.
sub  _validate_body_table
 {
  my $pll=shift;
  my $body = ($pll->Content( 'body' ))[0];

  my @content = $body->Content( '*' );
  my @rows    = $body->Content( 'row' );

  if(!$body->Attrib('cols') and !$body->Attrib('rows'))
   {
    return unless @rows;
    my $numrows = scalar @rows;
    die "$numrows row(s) found without 'rows' or 'cols' attributes.\n";
   }
  elsif($body->Attrib('cols') and !$body->Attrib('rows'))
   {
    my $numrows = scalar @rows;
    die "No 'rows' attribute specified, $numrows row(s) found.\n" if $numrows;
    die "In order to have 'cols', you must have 'rows'.\n";
   }
  elsif(@rows and grep { 'row' ne $_->Type() } @content)
   {
    die "Cannot use row elements with other child elements.\n";
   }
  elsif(scalar @rows != $body->Attrib('rows'))
   {
    die "Number of rows does not match \'rows\' attribute.\n";
   }
  else
   {
    my @grid;

    my ($i,$j);
    for($i=0;$i < $body->Attrib('rows');++$i)
     {
      my $r = $rows[$i];

      $j = 0;
      foreach my $c ($r->Content( '*' ))
       {
        my $type = $c->Type();  # Type is tested elsewhere

        ++$j while($grid[$i]->[$j]);
        
        foreach (0 .. ($c->Attrib('colspan') || 1) - 1)
         {
          foreach my $y (0 .. ($c->Attrib('rowspan') || 1) - 1)
           {
            $grid[$i+$y]->[$j] = 1;
           }
          $j++;
         }
       }

      die "Row $i number of columns does not match 'cols' attribute.\n"
           if !defined $grid[$i] or ($body->Attrib('cols')||0) != scalar @{$grid[$i]};
                                     #  The above weirdness deals with undef
     }

    die "cell extended below last row.\n" if @grid > $body->Attrib('rows');
   }
 }



#----------------------------------------------------------
#  convert a pll element to an appropriate output form.
sub   convert
 {
  my $pll = $_[0];

  my $type = $pll->Type();

  $LegalTags{$type}->{exec}->( @_ );
 }


#----------------------------------------------------------
#  Generate output for the layout element
sub   perform_layout
 {
  my $pll      = shift;
  my $page     = shift;
  my $wsdl     = shift;
  my $rContext = shift;
  my $output   = '';

  my $pre = ($pll->Content( 'prelayout' ))[0];
  $output .= ::resolve_macro( $pre->TextContent() ) if defined $pre;
  
  $output .= "<html>\n";

  foreach my $child ($pll->Content( '*' ))
   {
    $output .= convert( $child, $page, $wsdl, $rContext );
   }

  $output . "</html>\n";
 }


#----------------------------------------------------------
#  Generate output for the head element
sub   layout_head
 {
  my $head     = shift;
  my $page     = shift;
  my $wsdl     = shift;
  my $rContext = shift;
  
  my $output = "<head>\n<title>".$page->Attrib('title')."</title>\n";

  $output .= ::resolve_macro( "keywords" );
  $output .= ::resolve_macro( "description" );
  $output .= ::resolve_macro( "stylesheet" );

  foreach my $child ($head->Content( '*' ))
   {
    $output .= convert( $child, $page, $wsdl, $rContext );
   }
  
  $output . "</head>\n";
 }


#----------------------------------------------------------
#  Generate output for the body element
sub   layout_body
 {
  my $body     = shift;
  my $page     = shift;
  my $wsdl     = shift;
  my $rContext = shift;
  
  my $output;

  my $ba = ::resolve_macro( "body" );
  if($ba)
   {
    $output .= qq{<body $ba>\n};
   }
  else
   {
    $output .= qq{<body>\n};
   }

  if($body->HasAttrib('rows') or $body->HasAttrib('cols') or $body->Content( 'row' ))
   {
    $output .= qq{<table border="0" cellpadding="0" cellspacing="0">\n};
    foreach my $child ($body->Content( 'row' ))
     {
      $output .= convert( $child, $page, $wsdl, $rContext );
     }
    $output .= qq{</table>\n};
   }
  else
   {
    foreach my $child ($body->Content( '*' ))
     {
      $output .= convert( $child, $page, $wsdl, $rContext );
     }
   }
  
  $output . "</body>\n";
 }


 
#----------------------------------------------------------
#  Generate output for the row element
sub   layout_row
 {
  my $row      = shift;
  my $page     = shift;
  my $wsdl     = shift;
  my $rContext = shift;
  
  my $output = $row->Attrib('class')
               ? qq{<tr class="}.$row->Attrib('class').qq{">\n}
               : qq{<tr>\n};

  foreach my $child ($row->Content( '*' ))
   {
    $output .= convert( $child, $page, $wsdl, $rContext );
   }
  
  $output . "</tr>\n";
 }


#----------------------------------------------------------
#  Generate output for the cell element
sub   layout_cell
 {
  my $cell     = shift;
  my $page     = shift;
  my $wsdl     = shift;
  my $rContext = shift;
  
  my $output = join( ' ', qq{<td}, map { qq{$_="}.$cell->Attrib($_).qq{"} } 
                                   $cell->AttribNames() )
               . qq{>\n};

  foreach my $child ($cell->Content( '*' ))
   {
    $output .= convert( $child, $page, $wsdl, $rContext );
   }
  
  $output . "</td>\n";
 }

 
#----------------------------------------------------------
#  Generate output for the gutter element
sub   layout_gutter
 {
  my $gutter = shift;
  $gutter->{width} ||= DEFAULT_GUTTER_WIDTH;

  join( ' ', qq{<td}, map { qq{$_="}.$gutter->Attrib($_).qq{"} }
                      $gutter->AttribNames() )
               . qq{>&nbsp;</td>\n};
 }

 
#----------------------------------------------------------
#  Generate output for the code element
sub   layout_code
 {
  my $ref = $_[0]->{ref};

  my ($id, $args) = $ref =~ /^(\w+)(?:\(([^)]*)\))?/;

  $args = '' unless defined $args;

  $args = ::resolve_macros( $args );

  ::resolve_macro( "code[$id]($args)" );
 }

 
#----------------------------------------------------------
#  Generate output for several general elements that follow
#  the same form.
sub   layout_element
 {
  my $type = $_[0];
  my $ref = $_[1]->Attrib('ref');

  ::resolve_macro( defined $ref ? "$type"."[$ref]" : "$type" );
 }
 

 
#----------------------------------------------------------
#  Generate output for the content element
sub   layout_content
 {
  my $content  = shift;
  my $page     = shift;
  my $wsdl     = shift;
  my $rContext = shift;
  my $file     = $content->{file};

  return layout_element( 'content', $content, $page, $wsdl, $rContext )
                                                          unless defined $file;

  $file = buildfile( $rContext->[0]->Attrib('source'), $page->Attrib('src'), $file );
  my $output ="";

  local *F;
  local $/ = undef;

  open( F, $file ) or die "Unable to open \'$file\'.\n";
  $output = <F>;
  close( F );

  # Extract body of the HTML
  $output =~ s#^.*<[Bb][Oo][Dd][Yy][^>]*>##s;
  $output =~ s#</[Bb][Oo][Dd][Yy]>.*$##s;

  ::resolve_macros( $output );
 }
 
1;
